# creating a data frame version of the ICE-GB (heuristic version!) # ----------------------------------
rm(list=ls(all.names=TRUE)); invisible(gc()); invisible(library(car))

corpus.files <- dir("ICEGB", full.names=TRUE)
# set up big enough collector structures to insert results into
words <- tags <- mainverbs <- ditrans <- files <- rep(NA, 1300000)
starty <- 1; for(i in seq(corpus.files)) {
   current.corpus.file <- tolower(scan(corpus.files[i], what=character(), sep="\n", quiet=TRUE))
   # we find where the words are
   curr.words <- unlist(regmatches(current.corpus.file, gregexpr("(?<={).*(?=\\})", current.corpus.file, perl=TRUE)))
      curr.words <- gsub("<l>", "", curr.words, perl=TRUE)
   # we find where the tags are
   curr.tags <- regmatches(current.corpus.file, gregexpr("(?<=,)[a-z]+", current.corpus.file, perl=TRUE))
      curr.tags <- sapply(curr.tags, "[", 1)
   # we find where the main verbs are
   curr.mainverbs <- grepl("mvb,v", current.corpus.file, perl=TRUE)
   # we find where the ditransitives are
   curr.ditrans <- grepl("\\(ditr", current.corpus.file, perl=TRUE)
   # we store things
   endy <- starty + length(curr.words) - 1
   words[starty:endy]     <- curr.words    ; tags[starty:endy]      <- curr.tags
   mainverbs[starty:endy] <- curr.mainverbs; ditrans[starty:endy]   <- curr.ditrans
   files[starty:endy]     <- rep(substr(basename(corpus.files[i]),1,7), length(curr.words))
   starty <- endy+1
}
x <- data.frame(CASE=seq(words),FILE=factor(files), WORD=factor(words), TAG=factor(tags), MAINVERB=mainverbs, DITRANS=ditrans)
x <- droplevels(x[complete.cases(x),])
head(x)
#   CASE    FILE WORD      TAG MAINVERB DITRANS
# 1    1 S1A-001   ok      frm    FALSE   FALSE
# 2    2 S1A-001 adam        n    FALSE   FALSE
# 3    3 S1A-001  uhm interjec    FALSE   FALSE
# 4    4 S1A-001  <,>    pause    FALSE   FALSE
# 5    5 S1A-001 what     pron    FALSE   FALSE
# 6    6 S1A-001  did      aux    FALSE   FALSE
rm(list=setdiff(ls(), "x")); invisible(gc())



# observed frequencies & dispersions # ----------------------------------------------------------------
# requires a function KLD to compute DKLs
# setting up data structures
word.file.matrix <- table(x$WORD, x$FILE)
file.sizes.abs <- colSums(word.file.matrix)
file.sizes.rel <- file.sizes.abs/sum(file.sizes.abs)
# the frequencies of all words (kinda done already):
all.FREQS <- rowSums(word.file.matrix)
# the *DP*~norm~-values for all words:
smallest.file.size <- min(file.sizes.rel)
all.DPNORMS <- apply(word.file.matrix, 1, \(af) (sum(abs(af/sum(af)-file.sizes.rel))/2) / (1-smallest.file.size))
# the *KLD*~norm~-values for all words
all.KLDNORMS <- apply(word.file.matrix, 1, \(af) 1-(exp(-KLD(to.posterior.obs.data.p=af, from.prior.exp.theory.q=file.sizes.rel, inputs.r.freqs=TRUE, need.for.speed=TRUE))))
# add them to the original data frame
x$FREQ <- all.FREQS[x$WORD]
x$DPNORMS <- all.DPNORMS[x$WORD]
x$KLDNORMS <- all.KLDNORMS[x$WORD]
# check results so far
head(x)
#   CASE    FILE WORD      TAG MAINVERB DITRANS  FREQ   DPNORMS  KLDNORMS
# 1    1 S1A-001   ok      frm    FALSE   FALSE   567 0.7672705 0.9377118
# 2    2 S1A-001 adam        n    FALSE   FALSE    15 0.9909798 0.9987160
# 3    3 S1A-001  uhm interjec    FALSE   FALSE  5387 0.6215215 0.8123556
# 4    4 S1A-001  <,>    pause    FALSE   FALSE 20481 0.4874962 0.6687104
# 5    5 S1A-001 what     pron    FALSE   FALSE  3928 0.3675556 0.4578937
# 6    6 S1A-001  did      aux    FALSE   FALSE  1181 0.4767899 0.6575864


# CIs based on bag-of-words approach # ----------------------------------------------------------------
# returns the values given in the paper
sum(x$WORD=="give")
round(prop.test(444, nrow(x), 444/nrow(x))$conf.int * nrow(x), 1)
# 404.6, 487.3
sum(x$WORD=="given")
round(prop.test(365, nrow(x), 365/nrow(x))$conf.int * nrow(x), 1)
# 329.4, 404.4



# CIs based on boostrapping based on the bag-of-words approach # --------------------------------------
give.boot.binom <- rep(NA, 1000); weights <- prop.table(table(x$WORD=="give"))
set.seed(12); for (i in 1:1000) { give.boot.binom[i] <- mean(sample(x=0:1, size=nrow(x), replace=TRUE, prob=weights))*nrow(x) }
quantile(give.boot.binom, probs=c(0.025, 0.5, 0.975))
#    2.5%     50%   97.5%
# 405.000 445.000 484.025
given.boot.binom <- rep(NA, 1000); weights <- prop.table(table(x$WORD=="given"))
set.seed(12); for (i in 1:1000) { given.boot.binom[i] <- mean(sample(x=0:1, size=nrow(x), replace=TRUE, prob=weights))*nrow(x) }
quantile(given.boot.binom, probs=c(0.025, 0.5, 0.975))
# 2.5%   50% 97.5%
#  327   366   401



# Bootstrapping based on files (already for many lemmas and incl. freq & disp (KLD)) # ----------------
# verbs with an overall frequency between 100 and 10000
verbs.2.bootstrap <- table(paste0(x$WORD, "{", x$TAG))
verbs.2.bootstrap <- verbs.2.bootstrap[intersect(which(verbs.2.bootstrap>100), which(verbs.2.bootstrap<10000))]
verbs.2.bootstrap <- grep("{v$", names(verbs.2.bootstrap), perl=TRUE, value=TRUE)
verbs.2.bootstrap <- sort(sub("{v$", "", verbs.2.bootstrap, perl=TRUE)) # 179
lemmas.2.bootstrap <- c(
   "be","be","be","accept","agree","agree","allow","allow","be","be","ask","ask","base",
   "be","become","become","be","be","believe","bring","bring","build","buy","call","call",
   "come","carry","cause","change","change","come","come","come","concern","consider",
   "consider","cut","decide","describe","develop","do","do","do","do","expect","expect",
   "feel","feel","find","find","give","get","get","get","give","give","give","give","go",
   "go","go","go","get","have","happen","happen","happen","have","have","have","hear",
   "hear","hold","help","hope","include","involve","be","be","keep","know","know","know",
   "know","leave","leave","let","like","live","live","look","look","look","look","lose",
   "make","make","make","make","mean","mean","mean","meet","mention","meet","move","move",
   "need","need","pay","pay","play","play","produce","produce","provide","put","read",
   "remember","require","run","run","say","see","say","say","say","see","seem","see",
   "send","send","set","show","show","show","sit","speak","start","start","stop","suggest",
   "suppose","take","take","take","take","talk","talk","tell","thank","think","think",
   "think","tell","take","try","try","try","turn","turn","understand","use","use","use",
   "want","want","be","be","go","be","wish","work","work","work","write","write","write","write")
# define a number of iterations and collector structures:
iterations <- 1000; collector.freq <- collector.dpnorm <- collector.dpnofreq <- collector.kldnorm <-
   matrix(rep(NA, length(verbs.2.bootstrap)*iterations), ncol=iterations,
   dimnames=list(VERBS=verbs.2.bootstrap, ITERATIONS=seq(iterations)))
xx.w <- split(x$WORD, x$FILE); xx.w <- lapply(xx.w, as.character)
xx.f <- split(x$FILE, x$FILE); xx.f <- lapply(xx.f, as.character)
# bootstrap
set.seed(1); for (i in seq(iterations)) { cat(i, "\n")
   # determine which files are bootstrapped this time around
   files.bootstrapped <- sample(levels(x$FILE), 500, replace=TRUE)
   # brute-force sampling to not get the same file more than once in a row                             # !
   while (any(rle(files.bootstrapped)$lengths>1)) { files.bootstrapped <- sample(files.bootstrapped) } # !
   # creating the bootstrapped tdm
   ## get the words for the bootstrapped file names (as often as nec)
   bootstrapped.words <- unname(do.call(c, args=xx.w[files.bootstrapped]))
   ## get the file for the bootstrapped file names (as often as nec)
   bootstrapped.files <- unname(do.call(c, args=xx.f[files.bootstrapped]))
      bootstrapped.files <- rep(formatC(1:500, width=3, flag="0"), rle(bootstrapped.files)$length)
   ## create the temporary term-document matrix
   tdm.bootstrapped <- table(bootstrapped.words, bootstrapped.files)
   # compute this iteration's bootstrapped frequencies
   collector.freq[,i] <- rowSums(tdm.bootstrapped)[verbs.2.bootstrap]
   # compute this iteration's bootstrapped DPnorms
   filesizes.bootstrapped <- colSums(tdm.bootstrapped)/sum(tdm.bootstrapped) # exp
   smallest.file.size <- min(filesizes.bootstrapped)
   tdm.bootstrapped.red <- tdm.bootstrapped[verbs.2.bootstrap,]
   collector.dpnorm[,i] <- apply(tdm.bootstrapped.red, 1, \(af) (sum(abs(af/sum(af)-filesizes.bootstrapped))/2) / (1-smallest.file.size))
   # compute this iteration's bootstrapped KLD
   tdm.bootstrapped.rel <- prop.table(tdm.bootstrapped, 1)
   temp.KLDnorms <- apply(tdm.bootstrapped.rel, 1, \(af) KLD(to.posterior.obs.data.p=af, from.prior.exp.theory.q=filesizes.bootstrapped, inputs.r.freqs=TRUE, need.for.speed=TRUE))
   collector.kldnorm[,i] <- 1-exp(-temp.KLDnorms[verbs.2.bootstrap])
}



# CIs based on bootstrapping based on files # ---------------------------------------------------------
quantile(collector.freq["give",], probs=c(0.025, 0.5, 0.975))
# 2.5%   50% 97.5%
#  389   445   500
quantile(collector.freq["given",], probs=c(0.025, 0.5, 0.975))
# 2.5%   50% 97.5%
#  319   364   415



# various small checks/results # ----------------------------------------------------------------------
word.file.matrix[c("the", "try", "table", "tolstoy"), 1:3]
#         S1A-001 S1A-002 S1A-003
# the         113      42      61
# try           1       0       0
# table         0       0       0
# tolstoy       0       0       0

zipf.scale <- function (obs.freq, corpus.size.in.tokens) { 3+log10(obs.freq/corpus.size.in.tokens*1000000) }
all.FREQS <- rowSums(word.file.matrix)
   all.FREQS[c("the", "try", "table", "tolstoy")]
  #   the     try   table tolstoy
  # 58614     335      74       1
all.ZSFREQS <- zipf.scale(all.FREQS, sum(word.file.matrix))
   all.ZSFREQS[c("the", "try", "table", "tolstoy")]
#      the      try    table  tolstoy
# 7.738897 5.495941 4.840128 2.970896

head(file.sizes.abs, 3)
# S1A-001 S1A-002 S1A-003
#    2192    2162    2280
head(file.sizes.rel, 3)
#     S1A-001     S1A-002     S1A-003
# 0.002049918 0.002021863 0.002132214

all.KLDNORMS[c("the", "try", "table", "tolstoy")]
#       the       try     table   tolstoy
# 0.1011116 0.8020354 0.9772823 0.9998619

collector.freq[4:6,1:6]
#         ITERATIONS
# VERBS      1   2   3   4   5   6
#   accept 119 108  97  92 102 107
#   agree  160 171 155 172 147 146
#   agreed 135 116 114 107 132 107
